/* (C) 2022 YiRing, Inc. */
package com.yiring.app.job;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.xxl.job.core.handler.annotation.XxlJob;
import com.yiring.app.design.strategy.rule.ScanAlarmContext;
import com.yiring.app.domain.location.*;
import com.yiring.app.param.location.fence.LocationFenceJobParam;
import com.yiring.app.param.location.rule.RuleParam;
import com.yiring.app.service.location.fence.FenceAlarmService;
import com.yiring.app.vo.location.fence.LocationFenceJobVo;
import com.yiring.auth.domain.user.User;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author tml
 * @version 1.0
 * @date 2022/5/14 10:10
 */
@Slf4j
@Component
public class FenceAlarmJob {

    @Resource
    private LocationFenceRepository locationFenceRepository;

    @Resource
    private LocationFenceAlarmRepository locationFenceAlarmRepository;

    @Resource
    private LocationAlarmRuleRepository locationAlarmRuleRepository;

    @Resource
    private FenceAlarmPushLogRepository fenceAlarmPushLogRepository;

    @Resource
    private FenceAlarmService fenceAlarmService;

    @XxlJob("ScanAlarmHandler")
    @Transactional(rollbackFor = Exception.class)
    public void scanAlarmHandler() {
        //查询出所有已启用的电子围栏信息和正在报警的报警信息
        List<LocationFence> fenceList = locationFenceRepository.findAllByEnable(true);
        List<LocationFenceAlarm> alarmList = locationFenceAlarmRepository.findAllByStatus(
            LocationFenceAlarm.Status.ING
        );
        Map<Long, List<LocationFenceAlarm>> alarmMap = alarmList
            .stream()
            .collect(Collectors.groupingBy(e -> e.getFence().getId()));
        LocationFenceJobVo jobVo = LocationFenceJobVo
            .builder()
            .addAlarm(new ArrayList<>())
            .modifyAlarm(new ArrayList<>())
            .build();
        for (LocationFence fence : fenceList) {
            //获取该围栏下的所有报警规则
            Set<LocationFenceRule> fenceRuleSet = fence.getRules();
            for (LocationFenceRule fenceRule : fenceRuleSet) {
                RuleParam ruleParam = JSON.parseObject(fenceRule.getRule(), RuleParam.class);
                List<LocationFenceAlarm> oldAlarms = alarmMap.get(fence.getId());
                LocationFenceJobParam fenceJobParam = LocationFenceJobParam
                    .builder()
                    .rule(ruleParam)
                    .fenceId(fence.getId())
                    .fenceName(fence.getName())
                    .alarmTypeId(fenceRule.getAlarmType().getId())
                    .tagSet(fence.getTags())
                    .mapId(fence.getMapId())
                    .geometry(fence.getGeometry())
                    .oldAlarms(oldAlarms)
                    .build();
                LocationFenceJobVo locationFenceJobVo = ScanAlarmContext.scanAlarm(
                    fenceRule.getRelevanceParamType(),
                    fenceJobParam
                );
                jobVo.merge(locationFenceJobVo);
            }
        }
        //异步将新添加的报警信息推送给前端
        fenceAlarmService.alarmPushWeb(jobVo.getAddAlarm());
        //添加报警和关闭报警
        locationFenceAlarmRepository.saveAllAndFlush(jobVo.getAddAlarm());
        List<Long> ids = jobVo.getModifyAlarm().stream().map(LocationFenceAlarm::getId).collect(Collectors.toList());
        locationFenceAlarmRepository.batchOver(ids, LocalDateTime.now());
        log.info(
            "FenceAlarmJob.scanAlarmHandler: 添加的报警:[{}], 关闭的报警:[{}]",
            JSON.toJSONString(jobVo.getAddAlarm()),
            JSON.toJSONString(jobVo.getModifyAlarm())
        );

        //todo 推送报警信息

        //添加报警推送记录,可以考虑异步执行
        Set<Long> fenceIds = jobVo.getAddAlarm().stream().map(e -> e.getFence().getId()).collect(Collectors.toSet());
        List<LocationAlarmRule> locationAlarmRuleList = locationAlarmRuleRepository.findAllByFenceIds(fenceIds);
        List<FenceAlarmPushLog> pushLogList = new ArrayList<>();
        for (LocationFenceAlarm fenceAlarm : jobVo.getAddAlarm()) {
            //筛选出该条报警记录对应的报警规则
            locationAlarmRuleList =
                locationAlarmRuleList
                    .stream()
                    .filter(e -> fenceAlarm.getFence().getId().equals(e.getLocationFence().getId()))
                    .filter(e -> e.typesContains(fenceAlarm.getType()))
                    .collect(Collectors.toList());
            //一条报警记录可能对应多个报警规则(报警给谁的规则),报警规则的接收人和通知方式用笛卡尔积存到数据库
            for (LocationAlarmRule alarmRule : locationAlarmRuleList) {
                List<Integer> informManners = JSONArray.parseArray(alarmRule.getInformManner(), Integer.class);
                for (User user : alarmRule.getUsers()) {
                    for (Integer informManner : informManners) {
                        FenceAlarmPushLog pushLog = FenceAlarmPushLog
                            .builder()
                            .fenceAlarm(fenceAlarm)
                            .informManner(informManner)
                            .user(user)
                            .build();
                        pushLogList.add(pushLog);
                    }
                }
            }
        }
        //落库
        fenceAlarmPushLogRepository.saveAllAndFlush(pushLogList);
        log.info("FenceAlarmJob.scanAlarmHandler: 生成的推送记录:[{}]", JSON.toJSONString(pushLogList));
    }
}
