/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.risk.impl;

import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.github.liaochong.myexcel.core.SaxExcelReader;
import com.github.liaochong.myexcel.utils.StringUtil;
import com.yiring.app.constant.rehearsal.RiskLevelEnum;
import com.yiring.app.constant.risk.TemplateList;
import com.yiring.app.constant.risk.ThresholdStatusEnum;
import com.yiring.app.domain.risk.RiskWarn;
import com.yiring.app.domain.risk.RiskWarnRepository;
import com.yiring.app.domain.risk.Threshold;
import com.yiring.app.domain.workArea.WorkArea;
import com.yiring.app.domain.workArea.WorkAreaRepository;
import com.yiring.app.excel.risk.RiskWarnExcel;
import com.yiring.app.param.risk.RiskWarnAddParam;
import com.yiring.app.param.risk.RiskWarnConditionParam;
import com.yiring.app.param.risk.RiskWarnModifyParam;
import com.yiring.app.param.risk.ThresholdParam;
import com.yiring.app.service.risk.RiskWarnService;
import com.yiring.app.util.GeoUtils;
import com.yiring.app.util.JpaUtil;
import com.yiring.app.vo.risk.RiskWarnDetailVo;
import com.yiring.app.vo.risk.RiskWarnVo;
import com.yiring.common.constant.RegEx;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.param.IdParam;
import com.yiring.common.param.PageParam;
import com.yiring.common.vo.PageVo;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.Predicate;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.locationtech.jts.geom.Point;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

/**
 * @author tml
 * @version 1.0
 * @date 2022/5/23 17:57
 */
@Slf4j
@Service
@Transactional(rollbackFor = Exception.class)
public class RiskWarnServiceImpl implements RiskWarnService {

    @Resource
    private RiskWarnRepository riskWarnRepository;

    @Resource
    private WorkAreaRepository workAreaRepository;

    @Override
    public Result<String> addOne(RiskWarnAddParam param) {
        Result<String> result = checkThreshold(param.getThresholds());
        if (Status.OK.value() != result.getStatus()) {
            return result;
        }
        RiskWarn riskWarn = param.transform();
        riskWarnRepository.saveAndFlush(riskWarn);
        return Result.ok();
    }

    @Override
    public Result<String> modifyOne(RiskWarnModifyParam param) {
        Result<String> result = checkThreshold(param.getThresholds());
        if (Status.OK.value() != result.getStatus()) {
            return result;
        }
        Optional<RiskWarn> optional = riskWarnRepository.findById(param.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "修改的风险预警信息不存在");
        }
        RiskWarn riskWarn = optional.get();
        RiskWarn transform = param.transform();
        JpaUtil.copyNotNullProperties(transform, riskWarn);
        return Result.ok();
    }

    private Result<String> checkThreshold(List<ThresholdParam> thresholds) {
        boolean flag = thresholds.get(0).getMin().endsWith("%");
        for (ThresholdParam item : thresholds) {
            if (item.getMin().endsWith("%") != item.getMax().endsWith("%") || item.getMin().endsWith("%") != flag) {
                return Result.no(Status.EXPECTATION_FAILED, "阈值区间要么都为百分比,要么都不为百分比");
            }
            if (!item.getMin().matches(RegEx.DIGIT_DECIMALS) || !item.getMax().matches(RegEx.DIGIT_DECIMALS)) {
                return Result.no(Status.EXPECTATION_FAILED, "请输入正确格式的数字");
            }
        }
        return Result.ok();
    }

    @Override
    public Result<String> removeOne(IdParam idParam) {
        riskWarnRepository.deleteById(idParam.getId());
        return Result.ok();
    }

    @Override
    public Result<RiskWarnDetailVo> findOne(IdParam idParam) {
        Optional<RiskWarn> optional = riskWarnRepository.findById(idParam.getId());
        if (optional.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "查询的风险预警信息不存在");
        }
        RiskWarnDetailVo vo = RiskWarnDetailVo.transform(optional.get());
        return Result.ok(vo);
    }

    @Override
    public Result<PageVo<RiskWarnVo>> findList(RiskWarnConditionParam conditionParam, PageParam pageParam) {
        Pageable pageable = PageParam.toPageable(pageParam);
        Specification<RiskWarn> specification = getSpecification(conditionParam);
        Page<RiskWarn> page = riskWarnRepository.findAll(specification, pageable);
        List<RiskWarnVo> voList = page.get().map(RiskWarnVo::transform).collect(Collectors.toList());
        PageVo<RiskWarnVo> pageVo = PageVo.build(voList, voList.size());
        return Result.ok(pageVo);
    }

    @Override
    public void exportRiskWarn(RiskWarnConditionParam conditionParam, HttpServletResponse response) {
        Specification<RiskWarn> specification = getSpecification(conditionParam);
        List<RiskWarn> page = riskWarnRepository.findAll(specification);
        List<RiskWarnExcel> list = page.stream().map(RiskWarnExcel::transform).collect(Collectors.toList());
        try (
            DefaultStreamExcelBuilder<RiskWarnExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(RiskWarnExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(18)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(list);

            String fileName = URLEncoder.encode("风险预警信息表.xlsx", StandardCharsets.UTF_8);
            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("风险预警信息表导出失败: " + e.getMessage());
        }
    }

    @Override
    public Result<String> importRiskWarn(MultipartFile file) {
        if (file.isEmpty()) {
            return Result.no(Status.EXPECTATION_FAILED, "请选择一个excel文件");
        }
        List<RiskWarnExcel> riskWarnExcels;
        try (InputStream inputStream = file.getInputStream()) {
            riskWarnExcels =
                SaxExcelReader
                    .of(RiskWarnExcel.class)
                    .rowFilter(row -> row.getRowNum() > 0)
                    .ignoreBlankRow()
                    .read(inputStream);
        } catch (IOException e) {
            log.info(e.getMessage());
            throw new RuntimeException("文件导入异常");
        }
        //校验并转换数据
        Result<ArrayList<RiskWarn>> result = initImport(riskWarnExcels);
        if (Status.OK.value() != result.getStatus()) {
            return Result.no(Status.EXPECTATION_FAILED, result.getDetails());
        }
        List<RiskWarn> riskWarns = riskWarnRepository.saveAllAndFlush(result.getBody());
        return Result.ok("已成功添加" + riskWarns.size() + "添加");
    }

    /**
     * 校验并转换导入的预警信息
     * @param excels 导入的预警信息
     * @return 是否成功转换
     */
    private Result<ArrayList<RiskWarn>> initImport(List<RiskWarnExcel> excels) {
        //检查工区是否存在
        Set<String> set = excels.stream().map(RiskWarnExcel::getWorkAreaName).collect(Collectors.toSet());
        Specification<WorkArea> specification = (root, query, criteriaBuilder) -> {
            CriteriaBuilder.In<Object> in = criteriaBuilder.in(root.get(WorkArea.Fields.workAreaName));
            for (String item : set) {
                in.value(item);
            }
            return criteriaBuilder.and(in);
        };
        List<WorkArea> workAreaList = workAreaRepository.findAll(specification);
        Map<String, List<WorkArea>> map = workAreaList
            .stream()
            .collect(Collectors.groupingBy(WorkArea::getWorkAreaName));
        Set<String> stringSet = map.keySet();
        Set<String> fail = set.stream().filter(e -> !stringSet.contains(e)).collect(Collectors.toSet());
        if (fail.size() != 0) {
            return Result.no(Status.EXPECTATION_FAILED, "这些工区不存在:" + fail);
        }

        //检查位号是否重复
        Set<String> excelSet = excels.stream().map(RiskWarnExcel::getLocationNum).collect(Collectors.toSet());
        if (excelSet.size() != excels.size()) {
            return Result.no(Status.EXPECTATION_FAILED, "位号不能为空,并且不能重复");
        }
        Specification<RiskWarn> specifica = (root, query, criteriaBuilder) -> {
            CriteriaBuilder.In<Object> in = criteriaBuilder.in(root.get(RiskWarn.Fields.locationNum));
            for (String item : excelSet) {
                in.value(item);
            }
            return criteriaBuilder.and(in);
        };
        List<RiskWarn> riskWarnList = riskWarnRepository.findAll(specifica);
        if (riskWarnList.size() != 0) {
            Set<String> collect = riskWarnList.stream().map(RiskWarn::getLocationNum).collect(Collectors.toSet());
            return Result.no(Status.EXPECTATION_FAILED, "这些位号已经存在了:" + collect);
        }

        ArrayList<RiskWarn> list = new ArrayList<>();
        //每条数据以此检查
        for (RiskWarnExcel excel : excels) {
            if (StringUtil.isBlank(excel.getWorkAreaName())) {
                return Result.no(Status.EXPECTATION_FAILED, "工区不能为空");
            }
            if (StringUtil.isBlank(excel.getInstall())) {
                return Result.no(Status.EXPECTATION_FAILED, "安装位置不能为空");
            }
            if (StringUtil.isBlank(excel.getMonitoring())) {
                return Result.no(Status.EXPECTATION_FAILED, "监控参数不能为空");
            }
            WorkArea workArea = map.get(excel.getWorkAreaName()).get(0);
            //检查并转换量程和区间信息
            String[] array = { "", "" };
            boolean enable = true;
            ArrayList<Threshold> thresholds = new ArrayList<>();
            if (!StringUtil.isBlank(excel.getRange())) {
                array = excel.getRange().split("-");
                if (
                    array.length != 2 ||
                    !array[0].matches(RegEx.DIGIT_DECIMALS) ||
                    !array[1].matches(RegEx.DIGIT_DECIMALS)
                ) {
                    return Result.no(
                        Status.EXPECTATION_FAILED,
                        "里程格式" + excel.getRange() + "错误,正确格式是: 1-20"
                    );
                }
                Result<ArrayList<Threshold>> result = initThresholdS(excel, array);
                if (Status.OK.value() != result.getStatus()) {
                    return Result.no(Status.EXPECTATION_FAILED, result.getDetails());
                }
                thresholds = initThresholdS(excel, array).getBody();
            } else {
                enable = false;
            }
            String flame = excel.getFlame();
            if (!StringUtil.isBlank(flame)) {
                if (!checkThreshold(flame)) {
                    return Result.no(
                        Status.EXPECTATION_FAILED,
                        "报警值(熄灭)" + flame + "格式错误,正确格式: 1 或者 100%"
                    );
                }
                thresholds.add(
                    Threshold
                        .builder()
                        .min(flame)
                        .max(flame)
                        .status(ThresholdStatusEnum.FLAME_EXTINGUISH.getCode())
                        .build()
                );
            }
            //检查并转换坐标信息
            if (StringUtil.isBlank(excel.getPoint())) {
                return Result.no(Status.EXPECTATION_FAILED, "经纬度不能为空");
            }
            String[] split = excel.getPoint().split(",");
            if (split.length != 3) {
                return Result.no(
                    Status.EXPECTATION_FAILED,
                    "坐标格式" + excel.getPoint() + "错误,正确格式是: 112.863173604639,30.4786075211555,0"
                );
            }
            Point point = GeoUtils.createPoint(
                Double.parseDouble(split[0]),
                Double.parseDouble(split[1]),
                Double.parseDouble(split[2])
            );
            RiskWarn riskWarn = RiskWarn
                .builder()
                .workArea(workArea)
                .point(point)
                .install(excel.getInstall())
                .monitoring(excel.getMonitoring())
                .locationNum(excel.getLocationNum())
                .min(("").equals(array[0]) ? null : new BigDecimal(array[0]))
                .max(("").equals(array[1]) ? null : new BigDecimal(array[0]))
                .unit(excel.getUnit())
                .enable(enable)
                .thresholds(thresholds)
                .build();
            list.add(riskWarn);
        }
        return Result.ok(list);
    }

    /**
     * 将阈值区间转换
     * @param excel 阈值区间信息
     * @param array 量程
     * @return 转换后的阈值区间西悉尼
     */
    private Result<ArrayList<Threshold>> initThresholdS(RiskWarnExcel excel, String[] array) {
        ArrayList<Threshold> threshold = new ArrayList<>();
        Boolean flag = null;
        BigDecimal lowLowInt = null;
        BigDecimal lowInt = null;
        BigDecimal highInt = null;
        BigDecimal highHighInt = null;
        ArrayList<BigDecimal> list = new ArrayList<>();
        list.add(new BigDecimal(array[1]));

        String highHigh = excel.getHighHigh();
        if (!StringUtil.isBlank(highHigh)) {
            if (!checkThreshold(highHigh)) {
                return Result.no(
                    Status.EXPECTATION_FAILED,
                    "报警值(高高)" + highHigh + "格式错误,正确格式: 40 或者 90%"
                );
            }
            flag = highHigh.endsWith("%");

            if (flag) {
                highHighInt =
                    new BigDecimal(highHigh.substring(0, highHigh.length() - 1))
                        .divide(new BigDecimal(100), 4, BigDecimal.ROUND_HALF_UP);
            } else {
                highHighInt = new BigDecimal(highHigh);
            }
            list.add(highHighInt);
            BigDecimal max = list.get(list.size() - 2);
            String maxString = flag ? max.toString() + "%" : max.toString();
            threshold.add(
                Threshold
                    .builder()
                    .min(highHigh)
                    .max(maxString)
                    .status(ThresholdStatusEnum.HIGH_HIGH.getCode())
                    .riskLevel(RiskLevelEnum.RED.getCode())
                    .build()
            );
        }

        String high = excel.getHigh();
        if (!StringUtil.isBlank(high)) {
            if (!checkThreshold(high)) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值(高)" + high + "格式错误,正确格式: 30 或者 70%");
            }
            if (flag == null) {
                flag = high.endsWith("%");
            } else if (flag != high.endsWith("%")) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值要么都为百分比，要么都不为百分比");
            }
            if (flag) {
                highInt =
                    new BigDecimal(high.substring(0, high.length() - 1))
                        .divide(new BigDecimal(100), 4, BigDecimal.ROUND_HALF_UP);
            } else {
                highInt = new BigDecimal(high);
            }
            list.add(highInt);
            BigDecimal max = list.get(list.size() - 2);
            String maxString = flag ? max.toString() + "%" : max.toString();
            threshold.add(
                Threshold
                    .builder()
                    .min(high)
                    .max(maxString)
                    .status(ThresholdStatusEnum.HIGH.getCode())
                    .riskLevel(RiskLevelEnum.ORANGE.getCode())
                    .build()
            );
        }

        String low = excel.getLow();
        if (!StringUtil.isBlank(low)) {
            if (!checkThreshold(low)) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值(低)" + low + "格式错误,正确格式: 13 或者 30%");
            }
            if (flag == null) {
                flag = low.endsWith("%");
            } else if (flag != low.endsWith("%")) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值要么都为百分比，要么都不为百分比");
            }
            if (flag) {
                lowInt =
                    new BigDecimal(low.substring(0, low.length() - 1))
                        .divide(new BigDecimal(100), 4, BigDecimal.ROUND_HALF_UP);
            } else {
                lowInt = new BigDecimal(low);
            }
            list.add(lowInt);
            BigDecimal max = list.get(list.size() - 2);
            String maxString = flag ? max.toString() + "%" : max.toString();
            threshold.add(
                Threshold
                    .builder()
                    .min(low)
                    .max(maxString)
                    .status(ThresholdStatusEnum.LOW.getCode())
                    .riskLevel(RiskLevelEnum.YELLOW.getCode())
                    .build()
            );
        }

        String lowLow = excel.getLowLow();
        if (!StringUtil.isBlank(lowLow)) {
            if (!checkThreshold(lowLow)) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值(低低)" + lowLow + "格式错误,正确格式: 1 或者 15%");
            }
            if (flag == null) {
                flag = lowLow.endsWith("%");
            } else if (flag != lowLow.endsWith("%")) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值要么都为百分比，要么都不为百分比");
            }
            if (flag) {
                lowLowInt =
                    new BigDecimal(lowLow.substring(0, lowLow.length() - 1))
                        .divide(new BigDecimal(100), 4, BigDecimal.ROUND_HALF_UP);
            } else {
                lowLowInt = new BigDecimal(lowLow);
            }
            list.add(lowLowInt);
            BigDecimal max = list.get(list.size() - 2);
            String maxString = flag ? max.toString() + "%" : max.toString();
            threshold.add(
                Threshold
                    .builder()
                    .min(lowLow)
                    .max(maxString)
                    .status(ThresholdStatusEnum.LOW_LOW.getCode())
                    .riskLevel(RiskLevelEnum.BLUE.getCode())
                    .build()
            );
        }
        list.add(new BigDecimal(array[0]));
        if (flag != null && flag) {
            list.set(0, new BigDecimal(1));
            list.set(list.size() - 1, new BigDecimal(0));
        }
        for (int i = 1; i < list.size() - 1; i++) {
            if (list.get(i).compareTo(list.get(i + 1)) < 0) {
                return Result.no(Status.EXPECTATION_FAILED, "报警值区间值设置错误，风险高的值大于风险低的值了");
            }
        }
        return Result.ok(threshold);
    }

    /**
     * 判断字符串是否以%结尾或是数字或是小数
     * @param threshold 阈值
     * @return true:满足 false不满足
     */
    private boolean checkThreshold(String threshold) {
        return threshold.endsWith("%") || threshold.matches(RegEx.DIGIT_DECIMALS);
    }

    @Override
    public void exportTemplate(HttpServletResponse response) {
        List<RiskWarnExcel> list = TemplateList.getTemplateList();
        try (
            DefaultStreamExcelBuilder<RiskWarnExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(RiskWarnExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(18)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(list);

            String fileName = URLEncoder.encode("风险预警模板.xlsx", StandardCharsets.UTF_8);
            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("风险预警模板导出失败: " + e.getMessage());
        }
    }

    @Override
    public Result<String> batchEnable(Set<Long> ids, boolean enable) {
        int num = riskWarnRepository.batchEnable(ids, enable, LocalDateTime.now());
        return Result.ok((enable ? "启用了" : "停用了") + num + "条风险预警信息");
    }

    private Specification<RiskWarn> getSpecification(RiskWarnConditionParam conditionParam) {
        return (root, query, criteriaBuilder) -> {
            List<Predicate> list = new ArrayList<>();
            if (conditionParam.getWorkAreaId() != null) {
                list.add(criteriaBuilder.equal(root.get(RiskWarn.Fields.workArea), conditionParam.getWorkAreaId()));
            }
            if (conditionParam.getMonitoring() != null) {
                list.add(criteriaBuilder.equal(root.get(RiskWarn.Fields.monitoring), conditionParam.getMonitoring()));
            }
            Predicate[] array = list.toArray(new Predicate[0]);
            return criteriaBuilder.and(array);
        };
    }
}
