/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.warningMsg.impl;

import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.yiring.app.domain.warningMsg.WarningMessage;
import com.yiring.app.domain.warningMsg.WarningMessageRepository;
import com.yiring.app.domain.workArea.WorkArea;
import com.yiring.app.excel.warningMsg.WarningMessageExportExcel;
import com.yiring.app.param.warningMsg.WarningMessageParam;
import com.yiring.app.param.warningMsg.WarningMessageQueryParam;
import com.yiring.app.service.warningMsg.WarningMessageService;
import com.yiring.app.vo.warningMsg.WarningMessageCountVo;
import com.yiring.app.vo.warningMsg.WarningMessageRecVo;
import com.yiring.app.vo.warningMsg.WarningMessageVo;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.domain.BasicEntity;
import com.yiring.common.param.IdParam;
import com.yiring.common.param.IndexParam;
import com.yiring.common.vo.PageVo;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.persistence.EntityManager;
import javax.persistence.criteria.*;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.springframework.beans.BeanUtils;
import org.springframework.data.domain.*;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author tzl
 * @version 1.0
 * @description:
 * @date 2022/5/23 14:08
 */
@Slf4j
@Transactional(rollbackFor = RuntimeException.class)
@Service
public class WarningMessageServiceImpl implements WarningMessageService {

    @Resource
    WarningMessageRepository warningMessageRepository;

    @Resource
    EntityManager em;

    @Override
    public Result<String> update(WarningMessageParam warningMessageParam, IdParam idParam) {
        Optional<WarningMessage> byId = warningMessageRepository.findById(idParam.getId());
        if (byId.isEmpty()) {
            return Result.no(Status.NOT_FOUND);
        }
        WarningMessage warningMessage = byId.get();
        BeanUtils.copyProperties(warningMessageParam, warningMessage);
        //校验通知内容是否符合阿里云短信模板内容

        warningMessageRepository.saveAndFlush(warningMessage);
        return Result.ok();
    }

    @Override
    public Result<String> delete(IdParam idParam) {
        Optional<WarningMessage> byId = warningMessageRepository.findById(idParam.getId());
        if (byId.isEmpty()) {
            return Result.no(Status.NOT_FOUND);
        }
        WarningMessage warningMessage = byId.get();
        warningMessageRepository.delete(warningMessage);
        return Result.ok();
    }

    @Override
    public Result<PageVo<WarningMessageVo>> page(WarningMessageQueryParam warningMessageQueryParam) {
        PageVo<WarningMessageVo> resultVo;
        //排序
        Sort sort = Sort.by(Sort.Order.desc(BasicEntity.Fields.updateTime));
        if (
            Objects.nonNull(warningMessageQueryParam.getPageNo()) &&
            Objects.nonNull(warningMessageQueryParam.getPageSize())
        ) {
            //分页
            Pageable pageable = PageRequest.of(
                warningMessageQueryParam.getPageNo() - 1,
                warningMessageQueryParam.getPageSize()
            );
            Page<WarningMessage> all = warningMessageRepository.findAll(condition(warningMessageQueryParam), pageable);
            List<WarningMessageVo> data = all
                .get()
                .map(warningMessage -> {
                    WarningMessageVo vo = new WarningMessageVo();
                    BeanUtils.copyProperties(warningMessage, vo);
                    return vo;
                })
                .collect(Collectors.toList());
            resultVo = PageVo.build(data, all.getTotalElements());
        } else {
            List<WarningMessage> all = warningMessageRepository.findAll(condition(warningMessageQueryParam), sort);
            List<WarningMessageVo> data = all
                .stream()
                .map(warningMessage -> {
                    WarningMessageVo vo = new WarningMessageVo();
                    BeanUtils.copyProperties(warningMessage, vo);
                    return vo;
                })
                .collect(Collectors.toList());
            resultVo = PageVo.build(data, data.size());
        }
        return Result.ok(resultVo);
    }

    @Override
    public void export(WarningMessageQueryParam warningMessageQueryParam, HttpServletResponse httpServletResponse) {
        List<WarningMessage> workAreas = warningMessageRepository.findAll(condition(warningMessageQueryParam));
        List<WarningMessageExportExcel> warningMessageExportExcels = workAreas
            .stream()
            .map(warningMessage -> {
                WarningMessageExportExcel warningMessageExportExcel = new WarningMessageExportExcel();
                BeanUtils.copyProperties(warningMessage, warningMessageExportExcel);
                warningMessageExportExcel.setWorkAreaName(warningMessage.getWorkArea().getWorkAreaName());
                warningMessageExportExcel.setPersonInCharge(warningMessage.getWorkArea().getPersonInCharge());
                warningMessageExportExcel.setVideoName(warningMessage.getWorkArea().getVideo().getVideoName());
                return warningMessageExportExcel;
            })
            .toList();

        try (
            DefaultStreamExcelBuilder<WarningMessageExportExcel> defaultStreamExcelBuilder = DefaultStreamExcelBuilder
                .of(WarningMessageExportExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(14)
                .widths(8)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            defaultStreamExcelBuilder.append(warningMessageExportExcels);

            String fileName = URLEncoder.encode("预警信息.xlsx", StandardCharsets.UTF_8);
            httpServletResponse.setContentType("application/octet-stream");
            httpServletResponse.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            httpServletResponse.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = httpServletResponse.getOutputStream();

            Workbook workbook = defaultStreamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("导出预警信息失败: " + e.getMessage());
        }
    }

    public Specification<WarningMessage> condition(WarningMessageQueryParam warningMessageQueryParam) {
        return (root, query, criteriaBuilder) -> {
            List<Predicate> list = new ArrayList<>();
            Join<WarningMessage, WorkArea> join = root.join(WarningMessage.Fields.workArea, JoinType.LEFT);
            if (ObjectUtil.isNotNull(warningMessageQueryParam.getWorkAreaId())) {
                list.add(
                    criteriaBuilder.equal(join.get(BasicEntity.Fields.id), warningMessageQueryParam.getWorkAreaId())
                );
            }
            if (StrUtil.isNotBlank(warningMessageQueryParam.getTagNum())) {
                list.add(
                    criteriaBuilder.like(root.get(WarningMessage.Fields.tagNum), warningMessageQueryParam.getTagNum())
                );
            }
            if (StrUtil.isNotBlank(warningMessageQueryParam.getRiskLevel())) {
                list.add(
                    criteriaBuilder.equal(
                        root.get(WarningMessage.Fields.riskLevel),
                        warningMessageQueryParam.getRiskLevel()
                    )
                );
            }
            if (StrUtil.isNotBlank(warningMessageQueryParam.getThrState())) {
                list.add(
                    criteriaBuilder.equal(
                        root.get(WarningMessage.Fields.thrState),
                        warningMessageQueryParam.getThrState()
                    )
                );
            }
            if (StrUtil.isNotBlank(warningMessageQueryParam.getStatus())) {
                list.add(
                    criteriaBuilder.equal(root.get(WarningMessage.Fields.status), warningMessageQueryParam.getStatus())
                );
            }
            if (
                ObjectUtil.isNotNull(warningMessageQueryParam.getEndTime()) &&
                ObjectUtil.isNotNull(warningMessageQueryParam.getStartTime())
            ) {
                list.add(
                    criteriaBuilder.between(
                        root.get(BasicEntity.Fields.createTime),
                        warningMessageQueryParam.getStartTime(),
                        warningMessageQueryParam.getEndTime()
                    )
                );
            }
            Order order = criteriaBuilder.desc(root.get(BasicEntity.Fields.updateTime));
            Predicate[] predicates = new Predicate[list.size()];
            query.where(list.toArray(predicates));
            query.orderBy(order);
            return criteriaBuilder.and(list.toArray(predicates));
        };
    }

    @Override
    public Result<PageVo<WarningMessageCountVo>> list(WarningMessageQueryParam warningMessageQueryParam) {
        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<WarningMessageCountVo> cq = cb.createQuery(WarningMessageCountVo.class);
        Root<WarningMessage> root = cq.from(WarningMessage.class);
        Path<Object> objectPath = root.get(WarningMessage.Fields.riskLevel);
        Expression<Long> num = cb.count(root.get(BasicEntity.Fields.id));

        cq.multiselect(num, objectPath).groupBy(objectPath);

        List<Predicate> list = new ArrayList<>();
        if (
            ObjectUtil.isNotNull(warningMessageQueryParam.getStartTime()) &&
            ObjectUtil.isNotNull(warningMessageQueryParam.getEndTime())
        ) {
            list.add(
                cb.between(
                    root.get(BasicEntity.Fields.createTime),
                    warningMessageQueryParam.getStartTime(),
                    warningMessageQueryParam.getEndTime()
                )
            );
        }
        cq.where(list.toArray(new Predicate[0]));
        List<WarningMessageCountVo> resultList = em.createQuery(cq).getResultList();
        PageVo<WarningMessageCountVo> build = PageVo.build(resultList, resultList.size());
        return Result.ok(build);
    }

    @Override
    public Result<PageVo<WarningMessageRecVo>> listRec(IndexParam indexParam) {
        List<WarningMessage> all = warningMessageRepository.findAll(
            Example.of(WarningMessage.builder().tagNum(indexParam.getStr()).build())
        );
        List<WarningMessageRecVo> list = all
            .stream()
            .map(warningMessage -> {
                WarningMessageRecVo warningMessageRecVo = new WarningMessageRecVo();
                BeanUtils.copyProperties(warningMessage, warningMessageRecVo);
                return warningMessageRecVo;
            })
            .collect(Collectors.toList());
        PageVo<WarningMessageRecVo> build = PageVo.build(list, list.size());
        return Result.ok(build);
    }
    //    @Override
    //    public void exportRec(WarningMessageQueryParam warningMessageQueryParam, HttpServletResponse httpServletResponse) {
    //        List<WarningMessage> workAreas = warningMessageRepository.findAll(condition(warningMessageQueryParam));
    //        List<WarningMessageExportExcel> warningMessageExportExcels = workAreas
    //            .stream()
    //            .map(warningMessage -> {
    //                WarningMessageExportExcel warningMessageExportExcel = new WarningMessageExportExcel();
    //                BeanUtils.copyProperties(warningMessage, warningMessageExportExcel);
    //                warningMessageExportExcel.setWorkAreaName(warningMessage.getWorkArea().getWorkAreaName());
    //                warningMessageExportExcel.setPersonInCharge(warningMessage.getWorkArea().getPersonInCharge());
    //                warningMessageExportExcel.setVideoName(warningMessage.getWorkArea().getVideo().getVideoName());
    //                return warningMessageExportExcel;
    //            })
    //            .toList();
    //
    //        try (
    //            DefaultStreamExcelBuilder<WarningMessageExportExcel> defaultStreamExcelBuilder = DefaultStreamExcelBuilder
    //                .of(WarningMessageExportExcel.class)
    //                .threadPool(Executors.newFixedThreadPool(2))
    //                .rowHeight(14)
    //                .titleRowHeight(14)
    //                .widths(8)
    //                .style(
    //                    "cell->vertical-align:center;text-align:center",
    //                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
    //                )
    //                .start()
    //        ) {
    //            defaultStreamExcelBuilder.append(warningMessageExportExcels);
    //
    //            String fileName = URLEncoder.encode("预警信息.xlsx", StandardCharsets.UTF_8);
    //            httpServletResponse.setContentType("application/octet-stream");
    //            httpServletResponse.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
    //            httpServletResponse.setHeader("Content-Disposition", "attachment;filename=" + fileName);
    //            OutputStream out = httpServletResponse.getOutputStream();
    //
    //            Workbook workbook = defaultStreamExcelBuilder.fixedTitles().build();
    //            workbook.write(out);
    //            workbook.close();
    //            out.flush();
    //            out.close();
    //        } catch (Exception e) {
    //            log.error(e.getMessage(), e);
    //            throw new RuntimeException("导出预警信息失败: " + e.getMessage());
    //        }
    //    }
}
